/*
 * (C) Copyright 2000-2002
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * (C) Copyright 2001 Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Andreas Heppel <aheppel@sysgo.de>

 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/**************************************************************************
 *
 * Support for persistent environment data
 *
 * The "environment" is stored as a list of '\0' terminated
 * "name=value" strings. The end of the list is marked by a double
 * '\0'. New entries are always added at the end. Deleting an entry
 * shifts the remaining entries to the front. Replacing an entry is a
 * combination of deleting the old value and adding the new one.
 *
 * The environment is preceeded by a 32 bit CRC over the data part.
 *
 **************************************************************************
 */

#include <common.h>
#include <command.h>
#include <environment.h>
#include <watchdog.h>
#include <linux/stddef.h>
#include <asm/byteorder.h>
#include <macinfo.h>
#if (CONFIG_COMMANDS & CFG_CMD_NET)
#include <net.h>
#endif

#if !defined(CFG_ENV_IS_IN_NVRAM)	&& \
    !defined(CFG_ENV_IS_IN_EEPROM)	&& \
    !defined(CFG_ENV_IS_IN_FLASH)	&& \
    !defined(CFG_ENV_IS_IN_DATAFLASH)	&& \
    !defined(CFG_ENV_IS_NOWHERE)
# error Define one of CFG_ENV_IS_IN_{NVRAM|EEPROM|FLASH|DATAFLASH|NOWHERE}
#endif

#define XMK_STR(x)	#x
#define MK_STR(x)	XMK_STR(x)

extern ulong admin;
/************************************************************************
************************************************************************/

/* Function that returns a character from the environment */
extern char (*env_get_char)(int);

/* Function that returns a pointer to a value from the environment */
/* (Only memory version supported / needed). */
extern char *env_get_addr(int);

/* Function that updates CRC of the enironment */
extern void env_crc_update (void);

/************************************************************************
************************************************************************/

static int envmatch (char *, int);

/*
 * Table with supported baudrates (defined in config_xyz.h)
 */
static const unsigned long baudrate_table[] = CFG_BAUDRATE_TABLE;
#define	N_BAUDRATES (sizeof(baudrate_table) / sizeof(baudrate_table[0]))


/************************************************************************
 * Command interface: print one or all environment variables
 */

int do_printenv (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int i, j, k, nxt;
	int rcode = 0;

	if (argc == 1) {		/* Print all env variables	*/
		for (i=0; env_get_char(i) != '\0'; i=nxt+1) {
			for (nxt=i; env_get_char(nxt) != '\0'; ++nxt)
				;
			for (k=i; k<nxt; ++k)
				putc(env_get_char(k));
			putc  ('\n');

			if (ctrlc()) {
				puts ("\n ** Abort\n");
				return 1;
			}
		}

		printf("\nEnvironment size: %d/%d bytes\n", i, ENV_SIZE);

		return 0;
	}

	for (i=1; i<argc; ++i) {	/* print single env variables	*/
		char *name = argv[i];

		k = -1;

		for (j=0; env_get_char(j) != '\0'; j=nxt+1) {

			for (nxt=j; env_get_char(nxt) != '\0'; ++nxt)
				;
			k = envmatch(name, j);
			if (k < 0) {
				continue;
			}
			puts (name);
			putc ('=');
			while (k < nxt)
				putc(env_get_char(k++));
			putc ('\n');
			break;
		}
		if (k < 0) {
			printf ("## Error: \"%s\" not defined\n", name);
			rcode ++;
		}
	}
	return rcode;
}

/************************************************************************
 * Command interface: print one or all manufacturing data
 */
int do_printmd (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ManuInfo_t *manuinfo;
	uchar *flash_start;
	manuinfo = (uchar *)MANUFACTURING_RAM_START; 
	flash_start = (uchar *)MANUFACTURING_FLASH_START; 

	if((flash_start[0] == 0xFF) && (flash_start[1] == 0xFF)) {
		printf("manufacturing data is empty!\n");
		return 1;
	}
	printf("ProductID = %s\n", manuinfo->ProductID);
	printf("HWVer = %s\n", manuinfo->HWVer);
	printf("reginfo = %d\n", manuinfo->reginfo);
	printf("numofimages = %d\n", manuinfo->numofimages);
	printf("currimage = %d\n", manuinfo->currimage);
	printf("basemac = %s\n", manuinfo->basemac);
	printf("maccnt0 = %d\n", manuinfo->maccnt[0]);
	printf("maccnt1 = %d\n", manuinfo->maccnt[1]);
	printf("maccnt2 = %d\n", manuinfo->maccnt[2]);
	printf("maccnt3 = %d\n", manuinfo->maccnt[3]);
	printf("serno = %s\n", manuinfo->serno);
	printf("cpuinfo = %04x\n", manuinfo->boardinfo.cpuinfo);
	printf("freq = %d\n", manuinfo->boardinfo.freq);
	printf("ram = %d\n", manuinfo->boardinfo.ram);
	printf("flash = %d\n", manuinfo->boardinfo.flash);
	printf("poe_pwr = %d\n", manuinfo->boardinfo.poe_pwr);
	printf("portw1 = 0x%x\n", manuinfo->boardinfo.portw1);
	printf("portw2 = 0x%x\n", manuinfo->boardinfo.portw2);
	printf("portw3 = 0x%x\n", manuinfo->boardinfo.portw3);
	printf("portw4 = 0x%x\n", manuinfo->boardinfo.portw4);
	printf("portlan0 = 0x%x\n", manuinfo->boardinfo.portlan[0]);
	printf("portlan1 = 0x%x\n", manuinfo->boardinfo.portlan[1]);
	printf("portlan2 = 0x%x\n", manuinfo->boardinfo.portlan[2]);
	printf("portlan3 = 0x%x\n", manuinfo->boardinfo.portlan[3]);
	printf("portlan4 = 0x%x\n", manuinfo->boardinfo.portlan[4]);
	printf("portlan5 = 0x%x\n", manuinfo->boardinfo.portlan[5]);
	printf("portlan6 = 0x%x\n", manuinfo->boardinfo.portlan[6]);
	printf("portlan7 = 0x%x\n", manuinfo->boardinfo.portlan[7]);
	printf("minipci0 = 0x%x\n", manuinfo->boardinfo.minipci[0]);
	printf("minipci1 = 0x%x\n", manuinfo->boardinfo.minipci[1]);
	printf("serialport = %d\n", manuinfo->boardinfo.serialport);
	
	return 1;	
}


/************************************************************************
 * Set a new environment variable,
 * or replace or delete an existing one.
 *
 * This function will ONLY work with a in-RAM copy of the environment
 */

int do_setmd_back ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
//	char manuinfo_p[29][11] = {"ProductID", "reginfo", "HWVer", "numofimages", "currimage", "basemac", \
//		"maccnt0", "maccnt1", "maccnt2", "maccnt3", "serno", "cpuinfo", \
//		"freq", "ram", "flash", "poe_pwr", "portw1", "portw2", \
//		"portlan0", "portlan1", "portlan2", "portlan3", "portlan4", "portlan5", \
//		"portlan6", "portlan7", "minipci0", "minipci2", "serialport" };

	if(admin==0){
		if (input_passwd()==1)
			return 1;}
	ManuInfo_t *manuinfo;

	manuinfo = (uchar *)MANUFACTURING_RAM_START; 

	if(strcmp("ProductID", argv[1]) == 0) {
		memset(manuinfo->ProductID, 0x0, 32);
		sprintf(manuinfo->ProductID, "%s", argv[2]);
		manuinfo->ProductID[31] = 0x0;
	}
	else if(strcmp("HWVer", argv[1]) == 0) {
		memset(manuinfo->HWVer, 0x0, 12);
		sprintf(manuinfo->HWVer, "%s", argv[2]);
	}
	else if(strcmp("reginfo", argv[1]) == 0) {
		manuinfo->reginfo = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("numofimages", argv[1]) == 0) {
		manuinfo->numofimages = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("currimage", argv[1]) == 0) {
		manuinfo->currimage = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("basemac", argv[1]) == 0) {
		memset(manuinfo->basemac, 0x0, 12);
		sprintf(manuinfo->basemac, "%s", argv[2]);
	}
	else if(strcmp("maccnt0", argv[1]) == 0) {
		manuinfo->maccnt[0] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("maccnt1", argv[1]) == 0) {
		manuinfo->maccnt[1] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("maccnt2", argv[1]) == 0) {
		manuinfo->maccnt[2] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("maccnt3", argv[1]) == 0) {
		manuinfo->maccnt[3] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("serno", argv[1]) == 0) {
		memset(manuinfo->serno, 0x0, 32);
		sprintf(manuinfo->serno, "%s", argv[2]);
		manuinfo->serno[31] = 0x0;
	}
	else if(strcmp("cpuinfo", argv[1]) == 0) {
		manuinfo->boardinfo.cpuinfo = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("freq", argv[1]) == 0) {
		manuinfo->boardinfo.freq = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("ram", argv[1]) == 0) {
		manuinfo->boardinfo.ram = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("flash", argv[1]) == 0) {
		manuinfo->boardinfo.flash = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("poe_pwr", argv[1]) == 0) {
		manuinfo->boardinfo.poe_pwr = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("portw1", argv[1]) == 0) {
		manuinfo->boardinfo.portw1 = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("portw2", argv[1]) == 0) {
		manuinfo->boardinfo.portw2 = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("portlan0", argv[1]) == 0) {
		manuinfo->boardinfo.portlan[0] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("portlan1", argv[1]) == 0) {
		manuinfo->boardinfo.portlan[1] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("portlan2", argv[1]) == 0) {
		manuinfo->boardinfo.portlan[2] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("portlan3", argv[1]) == 0) {
		manuinfo->boardinfo.portlan[3] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("portlan4", argv[1]) == 0) {
		manuinfo->boardinfo.portlan[4] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("portlan5", argv[1]) == 0) {
		manuinfo->boardinfo.portlan[5] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("portlan6", argv[1]) == 0) {
		manuinfo->boardinfo.portlan[6] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("portlan7", argv[1]) == 0) {
		manuinfo->boardinfo.portlan[7] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("minipci0", argv[1]) == 0) {
		manuinfo->boardinfo.minipci[0] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("minipci1", argv[1]) == 0) {
		manuinfo->boardinfo.minipci[1] = simple_strtoul(argv[2], NULL, 10);
	}
	else if(strcmp("serialport", argv[1]) == 0) {
		manuinfo->boardinfo.serialport = simple_strtoul(argv[2], NULL, 10);
	}
	else {
		printf("Invalid Parameter!\n");
	}
	return 1;
}

int do_setmd ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
     	if(admin==0){
		if (input_passwd()==1)
			return 1;}

	char cmd[100];

	ManuInfo_t *manuinfo;

	manuinfo = (ManuInfo_t *)MANUFACTURING_RAM_START;
	memset(manuinfo,0,sizeof(ManuInfo_t));
	memcpy(manuinfo,(ulong *)(MANUFACTURING_FLASH_START), sizeof(ManuInfo_t));
		

	if(strcmp("basemac", argv[1]) == 0) {
                if(strlen(argv[2]) !=12 )
                {
                  printf("Mac address number should be 12 !\n");
                  return 1;
                }
		memset(manuinfo->basemac, 0x0, 12);
		sprintf(manuinfo->basemac, "%s", argv[2]);
	}
	
	else if (strcmp("serno", argv[1]) == 0) {
		memset(manuinfo->serno, 0x0, 32);
		sprintf(manuinfo->serno, "%s", argv[2]);
		manuinfo->serno[31] = 0x0;
	}
	else if (strcmp("productID", argv[1]) == 0) {
		memset(manuinfo->ProductID, 0x0, 32);
		sprintf(manuinfo->ProductID, "%s", argv[2]);
	}
	else 
	{
		printf ("Usage:setmd [basemac | serno | productID] [value]\n");
		return 1;
	}

	memset(cmd, 0x0, 100);
	sprintf(cmd, "protect off 0x%x 0x%x", MANUFACTURING_FLASH_START, MANUFACTURING_FLASH_START+CFG_MD_SIZE-1);
	run_command(cmd, 0);

	memset(cmd, 0x0, 100);
	sprintf(cmd, "erase 0x%x 0x%x", MANUFACTURING_FLASH_START, MANUFACTURING_FLASH_START+CFG_MD_SIZE-1);
	run_command(cmd, 0);
		
	memset(cmd, 0x0, 100);
	sprintf(cmd, "cp.b 0x%x 0x%x 0x%x", manuinfo, MANUFACTURING_FLASH_START,CFG_MD_SIZE);
	run_command(cmd, 0);
	
	return 0;
}
int do_getmd (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	
	if(admin==0){
		if (input_passwd()==1)
	return 1;}
				
	uchar *flash_start;
	flash_start = (uchar *)MANUFACTURING_FLASH_START; 

	ManuInfo_t *manuinfo;

	manuinfo = (ManuInfo_t *)MANUFACTURING_RAM_START;
	memset(manuinfo,0,sizeof(ManuInfo_t));
	memcpy(manuinfo,(ulong *)(MANUFACTURING_FLASH_START), sizeof(ManuInfo_t));

	if((flash_start[0] == 0xFF) && (flash_start[1] == 0xFF)) {
		printf("manufacturing data is empty!\n");
		return 1;
	}
        if(strcmp(argv[1],"basemac")==0)
		printf("basemac = %s\n", manuinfo->basemac);
	else if(strcmp(argv[1],"serno")==0)
		printf("serno = %s\n", manuinfo->serno);
	else if(strcmp(argv[1],"productID")==0)
		printf("productID = %s\n", manuinfo->ProductID);
	else
		{
			printf ("Usage:getmd [basemac | serno |productID]\n");
			return 1;
		}
	return 0;	
}
/************************************************************************
 * Set a new environment variable,
 * or replace or delete an existing one.
 *
 * This function will ONLY work with a in-RAM copy of the environment
 */

int _do_setenv (int flag, int argc, char *argv[])
{
	DECLARE_GLOBAL_DATA_PTR;

	int   i, len, oldval;
	int   console = -1;
	char *env, *nxt = NULL;
	char *name;
	bd_t *bd = gd->bd;

	char *env_data = env_get_addr(0);

	if (!env_data)	/* need copy in RAM */
		return 1;

	name = argv[1];


#if CONFIG_OCTEON
        /* Don't allow the loadaddr env variable to be changed if it is set based on
        ** a reserved block. */
	if (load_reserved_addr && strcmp(argv[1],"loadaddr") == 0)
        {
            printf("loadaddr not set due to load reserved memory allocated at 0x%x.\n", load_reserved_addr);
            printf("Please refer to the bootloader SDK documentation.\n");
            return 1;
        }
#endif


	/*
	 * search if variable with this name already exists
	 */
	oldval = -1;
	for (env=env_data; *env; env=nxt+1) {
		for (nxt=env; *nxt; ++nxt)
			;
		if ((oldval = envmatch(name, env-env_data)) >= 0)
			break;
	}

	/*
	 * Delete any existing definition
	 */
	if (oldval >= 0) {
#ifndef CONFIG_ENV_OVERWRITE

		/*
		 * Ethernet Address and serial# can be set only once,
		 * ver is readonly.
		 */
		if ( (strcmp (name, "serial#") == 0) ||
		    ((strcmp (name, "ethaddr") == 0)
#if defined(CONFIG_OVERWRITE_ETHADDR_ONCE) && defined(CONFIG_ETHADDR)
		     && (strcmp (env_get_addr(oldval),MK_STR(CONFIG_ETHADDR)) != 0)
#endif	/* CONFIG_OVERWRITE_ETHADDR_ONCE && CONFIG_ETHADDR */
		    ) ) {
			printf ("Can't overwrite \"%s\"\n", name);
			return 1;
		}
#endif

		/* Check for console redirection */
		if (strcmp(name,"stdin") == 0) {
			console = stdin;
		} else if (strcmp(name,"stdout") == 0) {
			console = stdout;
		} else if (strcmp(name,"stderr") == 0) {
			console = stderr;
		}

		if (console != -1) {
			if (argc < 3) {		/* Cannot delete it! */
				printf("Can't delete \"%s\"\n", name);
				return 1;
			}

			/* Try assigning specified device */
			if (console_assign (console, argv[2]) < 0)
				return 1;
		}

		/*
		 * Switch to new baudrate if new baudrate is supported
		 */
		if (strcmp(argv[1],"baudrate") == 0) {
			int baudrate = simple_strtoul(argv[2], NULL, 10);
			int i;
			for (i=0; i<N_BAUDRATES; ++i) {
				if (baudrate == baudrate_table[i])
					break;
			}
			if (i == N_BAUDRATES) {
				printf ("## Baudrate %d bps not supported\n",
					baudrate);
				return 1;
			}
			printf ("## Switch baudrate to %d bps and press ENTER ...\n",
				baudrate);
			udelay(50000);
			gd->baudrate = baudrate;
#ifdef CONFIG_PPC
			gd->bd->bi_baudrate = baudrate;
#endif

			serial_setbrg ();
			udelay(50000);
			for (;;) {
				if (getc() == '\r')
				      break;
			}
		}

		if (*++nxt == '\0') {
			if (env > env_data) {
				env--;
			} else {
				*env = '\0';
			}
		} else {
			for (;;) {
				*env = *nxt++;
				if ((*env == '\0') && (*nxt == '\0'))
					break;
				++env;
			}
		}
		*++env = '\0';
	}

#ifdef CONFIG_NET_MULTI
	if (strncmp(name, "eth", 3) == 0) {
		char *end;
		int   num = simple_strtoul(name+3, &end, 10);

		if (strcmp(end, "addr") == 0) {
			eth_set_enetaddr(num, argv[2]);
		}
	}
#endif

#ifdef CONFIG_OCTEON
        /* Have change in console UART take effect immediately */
	if (strncmp(name, "console_uart", 3) == 0) {
            int uart = simple_strtoul(argv[2], NULL, 10);
            if (uart == 0 || uart == 1)
            {
                gd->console_uart = uart;
int octeon_uart_setup(int uart_index);
                octeon_uart_setup(uart);
            }
	}
#endif

	/* Delete only ? */
	if ((argc < 3) || argv[2] == NULL) {
		env_crc_update ();
		return 0;
	}

	/*
	 * Append new definition at the end
	 */
	for (env=env_data; *env || *(env+1); ++env)
		;
	if (env > env_data)
		++env;
	/*
	 * Overflow when:
	 * "name" + "=" + "val" +"\0\0"  > ENV_SIZE - (env-env_data)
	 */
	len = strlen(name) + 2;
	/* add '=' for first arg, ' ' for all others */
	for (i=2; i<argc; ++i) {
		len += strlen(argv[i]) + 1;
	}
	if (len > (&env_data[ENV_SIZE]-env)) {
		printf ("## Error: environment overflow, \"%s\" deleted\n", name);
		return 1;
	}
	while ((*env = *name++) != '\0')
		env++;
	for (i=2; i<argc; ++i) {
		char *val = argv[i];

		*env = (i==2) ? '=' : ' ';
		while ((*++env = *val++) != '\0')
			;
	}

	/* end is marked with double '\0' */
	*++env = '\0';

	/* Update CRC */
	env_crc_update ();

	/*
	 * Some variables should be updated when the corresponding
	 * entry in the enviornment is changed
	 */

	if (strcmp(argv[1],"ethaddr") == 0) {
		char *s = argv[2];	/* always use only one arg */
		char *e;
		for (i=0; i<6; ++i) {
			bd->bi_enetaddr[i] = s ? simple_strtoul(s, &e, 16) : 0;
			if (s) s = (*e) ? e+1 : e;
		}
#ifdef CONFIG_NET_MULTI
		eth_set_enetaddr(0, argv[2]);
#endif
		return 0;
	}

	if (strcmp(argv[1],"ipaddr") == 0) {
		char *s = argv[2];	/* always use only one arg */
		char *e;
		unsigned long addr;
		bd->bi_ip_addr = 0;
		for (addr=0, i=0; i<4; ++i) {
			ulong val = s ? simple_strtoul(s, &e, 10) : 0;
			addr <<= 8;
			addr  |= (val & 0xFF);
			if (s) s = (*e) ? e+1 : e;
		}
		bd->bi_ip_addr = htonl(addr);
		return 0;
	}
	if (strcmp(argv[1],"loadaddr") == 0) {
		load_addr = simple_strtoul(argv[2], NULL, 16);
		return 0;
	}
#if (CONFIG_COMMANDS & CFG_CMD_NET)
	if (strcmp(argv[1],"bootfile") == 0) {
		copy_filename (BootFile, argv[2], sizeof(BootFile));
		return 0;
	}
#endif	/* CFG_CMD_NET */

#ifdef CONFIG_AMIGAONEG3SE
	if (strcmp(argv[1], "vga_fg_color") == 0 ||
	    strcmp(argv[1], "vga_bg_color") == 0 ) {
		extern void video_set_color(unsigned char attr);
		extern unsigned char video_get_attr(void);

		video_set_color(video_get_attr());
		return 0;
	}
#endif	/* CONFIG_AMIGAONEG3SE */

	return 0;
}

void setenv (char *varname, char *varvalue)
{
	char *argv[4] = { "setenv", varname, varvalue, NULL };
	_do_setenv (0, 3, argv);
}

int do_setenv ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	if (argc < 2) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	return _do_setenv (flag, argc, argv);
}

/************************************************************************
 * Prompt for environment variable
 */

#if (CONFIG_COMMANDS & CFG_CMD_ASKENV)
int do_askenv ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	if(admin==0){
		if (input_passwd()==1)
			return 1;}
	extern char console_buffer[CFG_CBSIZE];
	char message[CFG_CBSIZE];
	int size = CFG_CBSIZE - 1;
	int len;
	char *local_args[4];

	local_args[0] = argv[0];
	local_args[1] = argv[1];
	local_args[2] = NULL;
	local_args[3] = NULL;

	if (argc < 2) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}
	/* Check the syntax */
	switch (argc) {
	case 1:
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;

	case 2:		/* askenv envname */
		sprintf (message, "Please enter '%s':", argv[1]);
		break;

	case 3:		/* askenv envname size */
		sprintf (message, "Please enter '%s':", argv[1]);
		size = simple_strtoul (argv[2], NULL, 10);
		break;

	default:	/* askenv envname message1 ... messagen size */
	    {
		int i;
		int pos = 0;

		for (i = 2; i < argc - 1; i++) {
			if (pos) {
				message[pos++] = ' ';
			}
			strcpy (message+pos, argv[i]);
			pos += strlen(argv[i]);
		}
		message[pos] = '\0';
		size = simple_strtoul (argv[argc - 1], NULL, 10);
	    }
		break;
	}

	if (size >= CFG_CBSIZE)
		size = CFG_CBSIZE - 1;

	if (size <= 0)
		return 1;

	/* prompt for input */
	len = readline (message);

	if (size < len)
		console_buffer[size] = '\0';

	len = 2;
	if (console_buffer[0] != '\0') {
		local_args[2] = console_buffer;
		len = 3;
	}

	/* Continue calling setenv code */
	return _do_setenv (flag, len, local_args);
}
#endif	/* CFG_CMD_ASKENV */

/************************************************************************
 * Look up variable from environment,
 * return address of storage for that variable,
 * or NULL if not found
 */

char *getenv (char *name)
{
	int i, nxt;

	WATCHDOG_RESET();

	for (i=0; env_get_char(i) != '\0'; i=nxt+1) {
		int val;

		for (nxt=i; env_get_char(nxt) != '\0'; ++nxt) {
			if (nxt >= CFG_ENV_SIZE) {
				return (NULL);
			}
		}
		if ((val=envmatch(name, i)) < 0)
			continue;
		return (env_get_addr(val));
	}

	return (NULL);
}

int getenv_r (char *name, char *buf, unsigned len)
{
	int i, nxt;

	for (i=0; env_get_char(i) != '\0'; i=nxt+1) {
		int val, n;

		for (nxt=i; env_get_char(nxt) != '\0'; ++nxt) {
			if (nxt >= CFG_ENV_SIZE) {
				return (-1);
			}
		}
		if ((val=envmatch(name, i)) < 0)
			continue;
		/* found; copy out */
		n = 0;
		while ((len > n++) && (*buf++ = env_get_char(val++)) != '\0')
			;
		if (len == n)
			*buf = '\0';
		return (n);
	}
	return (-1);
}

#if defined(CFG_ENV_IS_IN_NVRAM) || defined(CFG_ENV_IS_IN_EEPROM) || \
    ((CONFIG_COMMANDS & (CFG_CMD_ENV|CFG_CMD_FLASH)) == \
      (CFG_CMD_ENV|CFG_CMD_FLASH))
int do_saveenv (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
#if !CFG_ENV_IS_NOWHERE
    extern char * env_name_spec;
	printf ("Saving Environment to %s...\n", env_name_spec);

	return (saveenv() ? 1 : 0);
#else
    printf("Environment updates not supported\n");
    return(1);
#endif
}

int do_savemd (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	if(admin==0){
		if (input_passwd()==1)
			return 1;}
	char	cmd[100];

	memset(cmd, 0x0, 100);
	sprintf(cmd, "protect off 0x%x 0x%x", MANUFACTURING_FLASH_START, MANUFACTURING_FLASH_START+CFG_MD_SIZE-1);
	run_command(cmd, 0);

	memset(cmd, 0x0, 100);
	sprintf(cmd, "erase 0x%x 0x%x", MANUFACTURING_FLASH_START, MANUFACTURING_FLASH_START+CFG_MD_SIZE-1);
	run_command(cmd, 0);

	memset(cmd, 0x0, 100);
	sprintf(cmd, "cp.b 0x%x 0x%x 0x%x", MANUFACTURING_RAM_START, MANUFACTURING_FLASH_START,CFG_MD_SIZE);
	run_command(cmd, 0);
}


int do_return2factory ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{

	char cmd[100];
	
	run_command("run nuke_env",0);
	
	// if erase the MD, the factory information will be erased together
	//sprintf(cmd, "protect off %x %x", CFG_MD_ADDR, CFG_MD_ADDR+CFG_MD_SIZE-1);
	//run_command(cmd, 0);
	
	//sprintf(cmd, "erase %x %x", CFG_MD_ADDR, CFG_MD_ADDR+CFG_MD_SIZE-1);
	//run_command(cmd, 0);

	printf("U-boot has been reseted to default settings, please reboot the device........\n");

	return 0;
}

#endif


/************************************************************************
 * Match a name / name=value pair
 *
 * s1 is either a simple 'name', or a 'name=value' pair.
 * i2 is the environment index for a 'name2=value2' pair.
 * If the names match, return the index for the value2, else NULL.
 */

static int
envmatch (char *s1, int i2)
{

	while (*s1 == env_get_char(i2++))
		if (*s1++ == '=')
			return(i2);
	if (*s1 == '\0' && env_get_char(i2-1) == '=')
		return(i2);
	return(-1);
}


/**************************************************/

U_BOOT_CMD(
	printenv, CFG_MAXARGS, 1,	do_printenv,
	"printenv- print environment variables\n",
	"\n    - print values of all environment variables\n"
	"printenv name ...\n"
	"    - print value of environment variable 'name'\n"
);

U_BOOT_CMD(
	setenv, CFG_MAXARGS, 0,	do_setenv,
	"setenv  - set environment variables\n",
	"name value ...\n"
	"    - set environment variable 'name' to 'value ...'\n"
	"setenv name\n"
	"    - delete environment variable 'name'\n"
);

U_BOOT_CMD(
	printmd, CFG_MAXARGS, 1, do_printmd,
	"",
	" - print values of all manufacturing data\n"
	"printmd name ...\n"
	"    - print value of manufacturing data 'name'\n"
	"parameter list: \n"
	"ProductID, HWVer, reginfo, numofimages, currimage, basemac,\n"
	"maccnt0, maccnt1, maccnt2, maccnt3, serno, cpuinfo, freq, ram,\n"
	"flash, poe_pwr, portw1, portw2, portlan0, portlan1, portlan2,\n"
	"portlan3, portlan4, portlan5, portlan6, portlan7, minipci0,\n"
	"minipci2, serialport\n"
);

/*U_BOOT_CMD(
	setmd, CFG_MAXARGS, 0,	do_setmd,
	"",
	" - set manufacturing data\n"
	"setmd name value ...\n"
	"    - set manufacturing data 'name' to 'value ...'\n"
	"parameter list: \n"
	"ProductID, HWVer, reginfo, numofimages, currimage, basemac,\n"
	"maccnt0, maccnt1, maccnt2, maccnt3, serno, cpuinfo, freq, ram,\n"
	"flash, poe_pwr, portw1, portw2, portlan0, portlan1, portlan2,\n"
	"portlan3, portlan4, portlan5, portlan6, portlan7, minipci0,\n"
	"minipci2, serialport\n"
);*/

U_BOOT_CMD(
	setmd, 3, 1,	do_setmd,
	"",
	" - set manufacturing data\n"
	"setmd name value ...\n"
	"    - set manufacturing data 'name' to 'value ...'\n"
	"parameter list: \n"
	"basemac,serno\n"
);
U_BOOT_CMD(
	getmd, 2, 1, do_getmd,
	"",
	" - print values of all manufacturing data\n"
	"printmd name ...\n"
	"    - print value of manufacturing data 'name'\n"
	"parameter list: \n"
	"basemac,serno\n"
);
#if ((CONFIG_COMMANDS & (CFG_CMD_ENV|CFG_CMD_FLASH)) == (CFG_CMD_ENV|CFG_CMD_FLASH))

U_BOOT_CMD(
	saveenv, 1, 0,	do_saveenv,
	"saveenv - save environment variables to persistent storage\n",
	NULL
);

U_BOOT_CMD(
	savemd, 1, 0,	do_savemd,
	"",
	"savemd - save manufacturing data to persistent storage\n"
);

U_BOOT_CMD(
	return2factory, CFG_MAXARGS, 0,	do_return2factory,
	"return2factory  - Reset whole bootloader configuration to factory settings.\n",
	""
);

#endif	/* CFG_CMD_ENV */

#if (CONFIG_COMMANDS & CFG_CMD_ASKENV)

U_BOOT_CMD(
	askenv,	CFG_MAXARGS,	1,	do_askenv,
	"askenv  - get environment variables from stdin\n",
	"name [message] [size]\n"
	"    - get environment variable 'name' from stdin (max 'size' chars)\n"
	"askenv name\n"
	"    - get environment variable 'name' from stdin\n"
	"askenv name size\n"
	"    - get environment variable 'name' from stdin (max 'size' chars)\n"
	"askenv name [message] size\n"
	"    - display 'message' string and get environment variable 'name'"
	"from stdin (max 'size' chars)\n"
);
#endif	/* CFG_CMD_ASKENV */

#if (CONFIG_COMMANDS & CFG_CMD_RUN)
int do_run (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
U_BOOT_CMD(
	run,	CFG_MAXARGS,	1,	do_run,
	"run     - run commands in an environment variable\n",
	"var [...]\n"
	"    - run the commands in the environment variable(s) 'var'\n"
);
#endif  /* CFG_CMD_RUN */
